@@ -5,54 +5,18 @@ import glob
 import os
 import re
 
-from trawl.path import path
-
-# These aren't all strictly necessary but I'm feeling too
-# lazy to separate the wheat from the chaffe right now.
-SPECIAL_METHODS = """
-repr str lt le eq ne gt ge cmp hash nonzero unicode call
-len getitem setitem delitem iter reversed contains getslice
-setslice delslice add sub mul floordiv mod divmod pow lshift
-rshift and xor or radd rsub rmul rdiv rtruediv rfloordiv rmod
-rdivmod rpow rlshift rrshift rand rxor ror iadd isub imul
-idiv itruediv ifloordiv imod ipow ilshift irshift iand ixor
-ior neg pos abs invert complex int long float oct hex index
-coerce enter exit
-""".split()
+from trawl.path import aspath, path
 
 def _exclude_cores(fn):
     return re.search(r"(^|[\/\\])core$", fn) and not os.path.isdir(fn)
 
-def skip_resolve(fn):
-    fn.__skip_resolve__ = True
-    return fn
-
-class FileListMeta(type):
-    def __new__(mcs, name, bases, dict):
-        ret = type.__new__(mcs, name, bases, dict)
-        for method in SPECIAL_METHODS:
-            base = getattr(list, "__%s__" % method, None)
-            if base is None:
-                continue
-            def mkfunc(func):
-                def _resolved(self, *args, **kwargs):
-                    self.resolve()
-                    return func(self, *args, **kwargs)
-                return _resolved
-            setattr(ret, "__%s__" % method, mkfunc(base))
-        return ret
-
-class FileListIterator(object):
-    def __init__(self, wrapped):
-        self.wrapped = wrapped
-
+class FileListIter(object):
+    def __init__(self, source):
+        self.source = source
     def next(self):
-        ret = self.wrapped.next()
-        if isinstance(ret, path):
-            return ret
-        return path(ret)
+        return aspath(self.source.next())
 
-class FileList(list):
+class FileList(object):
     """\
     A FileList is essentially an special class for dealing with lists of
     file names.
@@ -66,8 +30,6 @@ class FileList(list):
     is actually used.
     """
     
-    __metaclass__ = FileListMeta
-    
     DEFAULT_EXCLUDE_PATTERNS = [
         re.compile(r"(^|[\/\\])CVS([\/\\]|$)"),
         re.compile(r"(^|[\/\\])\.svn([\/\\]|$)"),
@@ -95,47 +57,210 @@ class FileList(list):
         """
         if len(patterns) == 1 and isinstance(patterns[0], FileList):
             other = patterns[0]
+            self._items = copy.copy(other._items)
             self._pending_add = copy.copy(other._pending_add)
             self._pending = other._pending
             self._exclude_patterns = copy.copy(other._exclude_patterns)
             self._exclude_funcs = copy.copy(other._exclude_funcs)
             self._exclude_regexps = copy.copy(other._exclude_regexps)
         else:
+            self._items = []
             self._pending_add = []
             self._pending = False
             self._exclude_patterns = FileList.DEFAULT_EXCLUDE_PATTERNS[:]
             self._exclude_funcs = FileList.DEFAULT_EXCLUDE_FUNCS[:]
             self._exclude_regexps = []
             map(self.include, patterns)
+    
+    # Not hashable
+    
+    __hash__ = None
+    
+    # Printing functions
+    
+    def __str__(self):
+        self.resolve()
+        return '[%s]' % ', '.join(map(str, self))
+    
+    def __repr__(self):
+        return "<FileList %s>" % str(self)
+    
+    # Binary operator methods
+    
+    def __add__(self, other):
+        print "Adding: %r" % other
+        if not isinstance(other, (list, FileList)):
+            return NotImplemented
+        elif isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            self._items += map(aspath, other._items)
+        else:
+            self.resolve()
+            self._items += map(aspath, other)
+        return self
+    
+    def __radd__(self, other):
+        if not isinstance(other, (list, FileList)):
+            return NotImplemented
+        elif isinstance(other, FileList):
+            self.resolve()
+            other.resolve
+            self._items = other._items + self._items
+        else:
+            self.resolve()
+            self._items = map(aspath, other) + self._items
+        return self
 
-    def __getattribute__(self, name):
-        """\
-        Make sure that any functions called on instances
-        return instances of FileList and not a plain list.
-        """
-        attr = super(FileList, self).__getattribute__(name)
-        if name == "__class__":
-            return attr
-        if callable(attr):
-            if not getattr(attr, "__skip_resolve__", False):
-                self.resolve()
-            @functools.wraps(attr)
-            def _wrap(*args, **kwargs):
-                ret = attr(*args, **kwargs)
-                if isinstance(ret, FileList):
-                    return ret
-                elif not isinstance(ret, list):
-                    return ret
-                return FileList(ret)
-            return _wrap
-        return attr
+    # Container Methods
+
+    def __len__(self):
+        self.resolve()
+        return len(self._items)
+
+    def __iter__(self):
+        self.resolve()
+        return FileListIter(iter(self._items))
+
+    def __getitem__(self, item):
+        self.resolve()
+        if not isinstance(item, slice):
+            return aspath(self._items[item])
+        ret = FileList(self)
+        ret._items = self._items[item]
+        return ret
+
+    def __setitem__(self, item, value):
+        self.resolve()
+        if not isinstance(item, slice):
+            self._items[item] = aspath(value)
+        elif isinstance(value, list):
+            self._items[item] = map(aspath, value)
+        elif isinstance(value, FileList):
+            value.resolve()
+            self._items[item] = value._items
+        else:
+            raise TypeError("%r is not an instance of list or FileList")
+
+    def __delitem__(self, item):
+        self.resolve()
+        del self._items[item]
+
+    def __contains__(self, item):
+        self.resolve()
+        return aspath(item) in self._items
+    
+    # Comparison functions
+    
+    def __lt__(self, other):
+        if isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            return self._items < other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items < other
+        return NotImplemented
+    
+    def __le__(self, other):
+        if isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            return self._items <= other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items <= other
+        return NotImplemented
+    
+    def __eq__(self, other):
+        if isinstance(other, FileList): 
+            self.resolve()
+            other.resolve()
+            return self._items == other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items == other
+        return NotImplemented
+    
+    def __ne__(self, other):
+        if isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            return self._items != other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items != other
+        return NotImplemented
+    
+    def __ge__(self, other):
+        if isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            return self._items >= other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items >= other
+        return NotImplemented
+    
+    def __gt__(self, other):
+        if isinstance(other, FileList):
+            self.resolve()
+            other.resolve()
+            return self._items >= other._items
+        elif isinstance(other, list):
+            self.resolve()
+            return self._items >= other
+        return NotImplemented
+
+    # List interface functions.
+
+    def append(self, obj):
+        self.resolve()
+        self._items.append(aspath(obj))
+    
+    def count(self, obj):
+        self.resolve()
+        return self._items.count(aspath(obj))
+    
+    def extend(self, iterobj):
+        self.resolve()
+        self._items.extend(aspath(p) for p in iterobj)
+    
+    def index(self, obj, *args):
+        self.resolve()
+        return self._items.index(aspath(obj), *args)
+
+    def insert(self, index, obj):
+        self.resolve()
+        self._items.insert(index, aspath(obj))
+
+    def pop(self, index):
+        self.resolve()
+        return aspath(self._items.pop(index))
+
+    def remove(self, obj):
+        self.resolve()
+        self._items.remove(aspath(obj))
+
+    def reverse(self):
+        self.resolve()
+        self._items.reverse()
+    
+    def sort(self, *args, **kwargs):
+        self.resolve()
+        self._items.sort(*args, **kwargs)
+
+    # FileList extension functions.
 
-    @skip_resolve
     def copy(self):
         "Return a cloned instance of this FileList"
         return self.__class__(self)
 
-    @skip_resolve
+    def expand(self):
+        ret = self.__class__(self)
+        ret.resolve()
+        return ret
+
     def include(self, *patterns):
         """\
         Add new glob patterns to this FileList instance. If an aray is given,
@@ -153,7 +278,6 @@ class FileList(list):
         self._pending = True
         return self
 
-    @skip_resolve
     def exclude(self, *patterns):
         """\
         Register a list of glob patterns that should be excluded from the
@@ -186,7 +310,6 @@ class FileList(list):
             self._resolve_excludes()
         return self
     
-    @skip_resolve
     def exclude_re(self, *patterns):
         """\
         Register a list of regular expression patterns or objects that
@@ -214,7 +337,14 @@ class FileList(list):
             self._resolve_excludes()
         return self
     
-    @skip_resolve
+    def is_excluded(self, fn):
+        "Should the given filename be excluded?"
+        if not self._exclude_regexps:
+            self._calculate_exclude_regexp()
+        if any(r.search(fn) for r in self._exclude_regexps):
+            return True
+        return any(func(fn) for func in self._exclude_funcs)
+    
     def clear_excludes(self):
         "Clear all the exclude patterns."
         self._exclude_patterns = []
@@ -223,7 +353,6 @@ class FileList(list):
             self._calculate_exclude_regexp()
         return self
 
-    @skip_resolve
     def resolve(self):
         "Resolve all the pending glob patterns."
         if self._pending:
@@ -240,17 +369,19 @@ class FileList(list):
 
         Triggers file name resolution.
         """
+        self.resolve()
         p = re.compile(pattern)
         for idx in range(len(self)):
             self[idx] = p.sub(repl, self[idx])
         return self
 
-    def existing(self):
+    def exist(self):
         """\
         Remove any file names that don't exist on the file system.
         
         Triggers file name resolution.
         """
+        self.resolve()
         idx = 0
         while idx < len(self):
             if not self[idx].exists():
@@ -259,20 +390,17 @@ class FileList(list):
                 idx += 1
         return self
 
-    @skip_resolve
     def _resolve_add(self, fn):
         if re.search(r"[*?\[\{]", fn):
             self._add_matching(fn)
         else:
-            self.append(fn)
+            self.append(aspath(fn))
 
-    @skip_resolve
     def _add_matching(self, pattern):
         "Add files matching the glob pattern."
-        fnames = filter(lambda fn: not self._exclude(fn), glob.glob(pattern))
-        self.extend(path(fn) for fn in fnames)
+        fnames = filter(lambda fn: not self.is_excluded(fn), glob.glob(pattern))
+        self.extend(aspath(fn) for fn in fnames)
 
-    @skip_resolve
     def _exclude(self, fn):
         "Should the given filename be excluded?"
         if not self._exclude_regexps:
@@ -281,17 +409,15 @@ class FileList(list):
             return True
         return any(func(fn) for func in self._exclude_funcs)
     
-    @skip_resolve
     def _resolve_excludes(self):
         self._calculate_exclude_regexp()
         idx = 0
         while idx < len(self):
-            if self._exclude(self[idx]):
+            if self.is_excluded(self[idx]):
                 self.pop(idx)
             else:
                 idx += 1
 
-    @skip_resolve
     def _calculate_exclude_regexp(self):
         self._exclude_regexps = []
         for e in self._exclude_patterns:
